
/*=============================================================================
   rtf2htm, a command-line program to convert RTF-formatted documents to HTML.
   Copyright (C) 2000,2001 Zachary Thayer Smith

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   The author is reachable by electronic mail at tuorfa@yahoo.com.
=============================================================================*/


/*----------------------------------------------------------------------
 * Module name:    word
 * Author name:    Zach Smith
 * Create date:    1 Sept 2000
 * Last change:    14 Oct 2000
 * Purpose:        Management of Word objects, which contain strings
 *                 as well as other Words.
 *----------------------------------------------------------------------
 * Changes:
 * 14 Oct 00, tuorfa@yahoo.com: fixed \fs bug (# is 2X the point size).
 * 14 Oct 00, tuorfa@yahoo.com: fixed table data printing.
 * 14 Oct 00, tuorfa@yahoo.com: protection against null entries in \info
 * 14 Oct 00, tuorfa@yahoo.com: fixed printing of <body> again
 * 14 Oct 00, tuorfa@yahoo.com: fixed closure of tables
 * 15 Oct 00, tuorfa@yahoo.com: fixed font attributes preceding <tr><td>
 * 15 Oct 00, tuorfa@yahoo.com: attributes now continue if >1 \cell in group
 * 15 Oct 00, tuorfa@yahoo.com: fixed font-size bug, lack of </head>
 *  7 Nov 00, tuorfa@yahoo.com: fixed \'## translatin bug
 *  8 Apr 01, tuorfa@yahoo.com: added check for out of memory after malloc
 * 21 Apr 01, tuorfa@yahoo.com: bug fixes regarding author, date
 * 21 Apr 01, tuorfa@yahoo.com: added paragraph alignment
 * 21 Apr 01, tuorfa@yahoo.com: fix for words getting lost after \par
 *--------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <malloc.h>
#include <string.h>

#include "defs.h"
#include "parse.h"
#include "main.h"
#include "error.h"
#include "word.h"
#include "hash.h"



/* Nested tables aren't supported.
 * Notice that <th> isn't here, since RTF doesn't have a direct equivalent.
 */
static int coming_pars_that_are_tabular = FALSE;
static int within_table = FALSE;
static int have_printed_tr=FALSE;
static int have_printed_td=FALSE;
static int have_printed_end_tr=FALSE;
static int have_printed_end_td=FALSE;


/* Paragraph alignment (kludge)
 */
#define ALIGN_LEFT 0
#define ALIGN_RIGHT 1
#define ALIGN_CENTER 2
#define ALIGN_JUSTIFY 3


/* Most pictures must be written to files. */
static int picture_file_number=1;


static int have_printed_body=FALSE;
static int within_head=TRUE;


#define CHARSET_ANSI 1
#define CHARSET_MAC 2
#define CHARSET_CP437 3
#define CHARSET_CP850 4
static int charset_type=CHARSET_ANSI;


/* local to getchar stuff */
int ungot_char=-1;
int ungot_char2=-1;


static char* special_exprs_ansi [] = {
/* 0x80 */ "","","","","","","","",
/* 0x88 */ "","","","","","","","",
/* 0x90 */ "","lsquo","rsquo","ldqyo","rdquo","bull","ndash","mdash",
/* 0x98 */ "","","","","","","","",
/* 0xa0 */ "nbsp","iexcl","cent","pound","curren","yen","brvbar","sect",
/* 0xa8 */ "uml","copy","ordf","laquo","not","shy","reg","macr",
/* 0xb0 */ "plus","plusmn","sup2","sup3","acute","micro","para","middot",
/* 0xb8 */ "cedil","sup1","frac14","ordm","raquo","frac12","frac34","iquest",
/* 0xc0 */ "Agrave","Aacute","Acirc","Atilde","Auml","Aring","AElig","Ccedil",
/* 0xc8 */ "Egrave","Eacute","Ecirc","Euml","Igrave","Iacute","Icirc","Iuml",
/* 0xd0 */ "ETH","Ntilde","Ograve","Oacute","Ocirc","Otilde","Ouml","times",
/* 0xd8 */ "Oslash","Ugrave","Uacute","Ucirc","Uuml","Yacute","THORN","szlig",
/* 0xe0 */ "agrave","aacute","acirc","atilde","auml","aring","aelig","ccedil",
/* 0xe8 */ "egrave","eacute","ecirc","euml","igrave","iacute","icirc","iuml",
/* 0xf0 */ "eth","ntilde","ograve","oacute","ocirc","otilde","ouml","divide",
/* 0xf8 */ "oslash","ugrave","uacute","ucirc","uuml","yacute","thorn","yuml",
};

static char* special_exprs_mac [] = {
/* 0x80 */ "","","","","","","","",
/* 0x88 */ "","","","","","","","",
/* 0x90 */ "","","","","","","","",
/* 0x98 */ "","","","","","","","",
/* 0xa0 */ "","","","","bull","","","",
/* 0xa8 */ "","","","","","","","",
/* 0xb0 */ "","","","","","","","",
/* 0xb8 */ "","","","","","","","",
/* 0xc0 */ "","","","","","","","",
/* 0xc8 */ "","","","","","","","",
/* 0xd0 */ "mdash","ndash","ldquo","rdquo","lquo","rquo","","",
/* 0xd8 */ "","","","","","","","",
/* 0xe0 */ "","","","","","","","",
/* 0xe8 */ "","","","","","","","",
/* 0xf0 */ "","","","","","","","",
/* 0xf8 */ "","","","","","","","",
};

static char* special_exprs_cp437 [] = {
/* 0x80 */ "ccedil",
/* 0x81 */ "uuml",
/* 0x82 */ "eacute",
/* 0x83 */ "acirc",
/* 0x84 */ "auml",
/* 0x85 */ "agrave",
/* 0x86 */ "aring",
/* 0x87 */ "ccedil",
/* 0x88 */ "ecirc",
/* 0x89 */ "euml",
/* 0x8a */ "egrave",
/* 0x8b */ "iuml",
/* 0x8c */ "icirc",
/* 0x8d */ "igrave",
/* 0x8e */ "auml",
/* 0x8f */ "aring",
/* 0x90 */ "eacute",
/* 0x91 */ "aelig",
/* 0x92 */ "aelig",
/* 0x93 */ "ocirc",
/* 0x94 */ "ouml",
/* 0x95 */ "ograve",
/* 0x96 */ "ucirc",
/* 0x97 */ "ugrave",
/* 0x98 */ "yuml",
/* 0x99 */ "ouml",
/* 0x9a */ "uuml",
/* 0x9b */ "cent",
/* 0x9c */ "pound",
/* 0x9d */ "yen",
/* 0x9e */ "#8359", /* peseta */
/* 0x9f */ "#402", /* small f with hook */
/* 0xa0 */ "aacute",
/* 0xa1 */ "iacute",
/* 0xa2 */ "oacute",
/* 0xa3 */ "uacute",
/* 0xa4 */ "ntilde",
/* 0xa5 */ "ntilde",
/* 0xa6 */ "ordf",
/* 0xa7 */ "frac14",
/* 0xa8 */ "iquest",
/* 0xa9 */ "#8976", /* reversed not */
/* 0xaa */ "not",
/* 0xab */ "frac12",
/* 0xac */ "raquo",
/* 0xad */ "iexcl",
/* 0xae */ "laquo",
/* 0xaf */ "ordm",
/* 0xb0 */ "#9617", /* light shade */
/* 0xb1 */ "#9618", /* med. shade */
/* 0xb2 */ "#9619", /* dark shade */
/* 0xb3 */ "#9474", /* box-draw light vert. */
/* 0xb4 */ "#9508", /* box-draw light vert. + lt. */
/* 0xb5 */ "#9569", /* box-draw vert. sgl. + lt. dbl. */
/* 0xb6 */ "#9570", /* box-draw vert. dbl. + lt. sgl. */
/* 0xb7 */ "#9558", /* box-draw dn. dbl. + lt. sgl. */
/* 0xb8 */ "#9557", /* box-draw dn. sgl. + lt. dbl. */
/* 0xb9 */ "#9571", /* box-draw dbl. vert. + lt. */
/* 0xba */ "#9553", /* box-draw dbl. vert. */
/* 0xbb */ "#9559", /* box-draw dbl. dn. + lt. */
/* 0xbc */ "#9565", /* box-draw dbl. up + lt. */
/* 0xbd */ "#9564", /* box-draw up dbl. + lt. sgl. */
/* 0xbe */ "#9563", /* box-draw up sgl. + lt. dbl. */
/* 0xbf */ "#9488", /* box-draw light dn. + lt. */
/* 0xc0 */ "#9492", /* box-draw light up + rt. */
/* 0xc1 */ "#9524", /* box-draw light up + horiz. */
/* 0xc2 */ "#9516", /* box-draw light dn. + horiz. */
/* 0xc3 */ "#9500", /* box-draw light vert. + rt. */
/* 0xc4 */ "#9472", /* box-draw light horiz. */
/* 0xc5 */ "#9532", /* box-draw light vert. + horiz. */
/* 0xc6 */ "#9566", /* box-draw vert. sgl. + rt. dbl. */
/* 0xc7 */ "#9567", /* box-draw vert. dbl. + rt. sgl. */
/* 0xc8 */ "#9562", /* box-draw dbl. up + rt. */
/* 0xc9 */ "#9556", /* box-draw dbl. dn. + rt. */
/* 0xca */ "#9577", /* box-draw dbl. up + horiz. */
/* 0xcb */ "#9574", /* box-draw dbl. dn. + horiz. */
/* 0xcc */ "#9568", /* box-draw dbl. vert. + rt. */
/* 0xcd */ "#9552", /* box-draw dbl. horiz. */
/* 0xce */ "#9580", /* box-draw dbl. vert. + horiz. */
/* 0xcf */ "#9575", /* box-draw up sgl. + horiz. dbl. */
/* 0xd0 */ "#9576", /* box-draw up dbl. + horiz. sgl. */
/* 0xd1 */ "#9572", /* box-draw dn. sgl. + horiz. dbl. */
/* 0xd2 */ "#9573", /* box-draw dn. dbl. + horiz. sgl. */
/* 0xd3 */ "#9561", /* box-draw up dbl. + rt. sgl. */
/* 0xd4 */ "#9560", /* box-draw up sgl. + rt. dbl. */
/* 0xd5 */ "#9554", /* box-draw dn. sgl. + rt. dbl. */
/* 0xd6 */ "#9555", /* box-draw dn. dbl. + rt. sgl. */
/* 0xd7 */ "#9579", /* box-draw vert. dbl. + horiz. sgl. */
/* 0xd8 */ "#9578", /* box-draw vert. sgl. + horiz. dbl. */
/* 0xd9 */ "#9496", /* box-draw light up + lt. */
/* 0xda */ "#9484", /* box-draw light dn. + rt. */
/* 0xdb */ "#9608", /* full block */
/* 0xdc */ "#9604", /* lower 1/2 block */
/* 0xdd */ "#9612", /* lt. 1/2 block */
/* 0xde */ "#9616", /* rt. 1/2 block */
/* 0xdf */ "#9600", /* upper 1/2 block */
/* 0xe0 */ "#945", /* greek small alpha */
/* 0xe1 */ "szlig",
/* 0xe2 */ "#915", /* greek cap gamma */
/* 0xe3 */ "#960", /* greek small pi */
/* 0xe4 */ "#931", /* greek cap sigma */
/* 0xe5 */ "#963", /* greek small sigma */
/* 0xe6 */ "micro",
/* 0xe7 */ "#964", /* greek small tau */
/* 0xe8 */ "#934", /* greek cap phi */
/* 0xe9 */ "#920", /* greek cap theta */
/* 0xea */ "#937", /* greek cap omega */
/* 0xeb */ "#948", /* greek small delta */
/* 0xec */ "#8734", /* inf. */
/* 0xed */ "#966", /* greek small phi */
/* 0xee */ "#949", /* greek small epsilon */
/* 0xef */ "#8745", /* intersect */
/* 0xf0 */ "#8801", /* identical */
/* 0xf1 */ "plusmn",
/* 0xf2 */ "#8805", /* greater-than or equal to */
/* 0xf3 */ "#8804", /* less-than or equal to */
/* 0xf4 */ "#8992", /* top 1/2 integral */
/* 0xf5 */ "#8993", /* bottom 1/2 integral */
/* 0xf6 */ "divide",
/* 0xf7 */ "#8776", /* almost = */
/* 0xf8 */ "plus",
/* 0xf9 */ "#8729", /* bullet op */
/* 0xfa */ "middot",
/* 0xfb */ "#8730", /* sqrt */
/* 0xfc */ "#8319", /* super-script small n */
/* 0xfd */ "sup2",
/* 0xfe */ "#9632", /* black square */
/* 0xff */ "nbsp",
};

static char* special_exprs_cp850 [] = {
/* 0x80 */  "ccedil",
/* 0x81 */  "uuml",
/* 0x82 */  "eacute",
/* 0x83 */  "acirc",
/* 0x84 */  "auml",
/* 0x85 */  "agrave",
/* 0x86 */  "aring",
/* 0x87 */  "ccedil",
/* 0x88 */  "ecirc",
/* 0x89 */  "euml",
/* 0x8a */  "egrave",
/* 0x8b */  "iuml",
/* 0x8c */  "icirc",
/* 0x8d */  "igrave",
/* 0x8e */  "auml",
/* 0x8f */  "aring",
/* 0x90 */  "eacute",
/* 0x91 */  "aelig",
/* 0x92 */  "aelig",
/* 0x93 */  "ocirc",
/* 0x94 */  "ouml",
/* 0x95 */  "ograve",
/* 0x96 */  "ucirc",
/* 0x97 */  "ugrave",
/* 0x98 */  "yuml",
/* 0x99 */  "ouml",
/* 0x9a */  "uuml",
/* 0x9b */  "oslash",
/* 0x9c */  "pound",
/* 0x9d */  "oslash",
/* 0x9e */  "times",
/* 0x9f */  "#402", /* small f with hook */
/* 0xa0 */  "aacute",
/* 0xa1 */  "iacute",
/* 0xa2 */  "oacute",
/* 0xa3 */  "uacute",
/* 0xa4 */  "ntilde",
/* 0xa5 */  "ntilde",
/* 0xa6 */  "ordf",
/* 0xa7 */  "frac14",
/* 0xa8 */  "iquest",
/* 0xa9 */  "reg",
/* 0xaa */  "not",
/* 0xab */  "frac12",
/* 0xac */  "raquo",
/* 0xad */  "iexcl",
/* 0xae */  "laquo",
/* 0xaf */  "ordm",
/* 0xb0 */  "#9617", /* light shade */
/* 0xb1 */  "#9618", /* med. shade */
/* 0xb2 */  "#9619", /* dark shade */
/* 0xb3 */  "#9474", /* box-draw light vert. */
/* 0xb4 */  "#9508", /* box-draw light vert. + lt. */
/* 0xb5 */  "aacute",
/* 0xb6 */  "acirc",
/* 0xb7 */  "agrave",
/* 0xb8 */  "copy",
/* 0xb9 */  "#9571", /* box-draw dbl. vert. + lt. */
/* 0xba */  "#9553", /* box-draw dbl. vert. */
/* 0xbb */  "#9559", /* box-draw dbl. dn. + lt. */
/* 0xbc */  "#9565", /* box-draw dbl. up + lt. */
/* 0xbd */  "cent",
/* 0xbe */  "yen",
/* 0xbf */  "#9488", /* box-draw light dn. + lt. */
/* 0xc0 */  "#9492", /* box-draw light up + rt. */
/* 0xc1 */  "#9524", /* box-draw light up + horiz. */
/* 0xc2 */  "#9516", /* box-draw light dn. + horiz. */
/* 0xc3 */  "#9500", /* box-draw light vert. + rt. */
/* 0xc4 */  "#9472", /* box-draw light horiz. */
/* 0xc5 */  "#9532", /* box-draw light vert. + horiz. */
/* 0xc6 */  "atilde",
/* 0xc7 */  "atilde",
/* 0xc8 */  "#9562", /* box-draw dbl. up + rt. */
/* 0xc9 */  "#9556", /* box-draw dbl. dn. + rt. */
/* 0xca */  "#9577", /* box-draw dbl. up + horiz. */
/* 0xcb */  "#9574", /* box-draw dbl. dn. + horiz. */
/* 0xcc */  "#9568", /* box-draw dbl. vert. + rt. */
/* 0xcd */  "#9552", /* box-draw dbl. horiz. */
/* 0xce */  "#9580", /* box-draw dbl. vert. + horiz. */
/* 0xcf */  "curren",
/* 0xd0 */  "eth",
/* 0xd1 */  "eth",
/* 0xd2 */  "ecirc",
/* 0xd3 */  "euml",
/* 0xd4 */  "egrave",
/* 0xd5 */  "#305", /* small dotless i */
/* 0xd6 */  "iacute",
/* 0xd7 */  "icirc",
/* 0xd8 */  "iuml",
/* 0xd9 */  "#9496", /* box-draw light up + lt. */
/* 0xda */  "#9484", /* box-draw light dn. + rt. */
/* 0xdb */  "#9608", /* full-block */
/* 0xdc */  "#9604", /* lower 1/2 block */
/* 0xdd */  "brvbar",
/* 0xde */  "igrave",
/* 0xdf */  "#9600", /* upper 1/2 block */
/* 0xe0 */  "oacute",
/* 0xe1 */  "szlig",
/* 0xe2 */  "ocirc",
/* 0xe3 */  "ograve",
/* 0xe4 */  "otilde",
/* 0xe5 */  "otilde",
/* 0xe6 */  "micro",
/* 0xe7 */  "thorn",
/* 0xe8 */  "thorn",
/* 0xe9 */  "uacute",
/* 0xea */  "ucirc",
/* 0xeb */  "ugrave",
/* 0xec */  "yacute",
/* 0xed */  "yacute",
/* 0xee */  "macr",
/* 0xef */  "acute",
/* 0xf0 */  "shy",
/* 0xf1 */  "plusmn",
/* 0xf2 */  "#8215", /* dbl. lowline */
/* 0xf3 */  "frac34",
/* 0xf4 */  "para",
/* 0xf5 */  "sect",
/* 0xf6 */  "divide",
/* 0xf7 */  "cedil",
/* 0xf8 */  "plus",
/* 0xf9 */  "uml",
/* 0xfa */  "middot",
/* 0xfb */  "sup1",
/* 0xfc */  "sup3",
/* 0xfd */  "sup2",
/* 0xfe */  "#9632", /* black square */
/* 0xff */  "nbsp",
};



void starting_body();
void starting_text();


static int banner_printed=FALSE;

void
print_banner () {
	if (!banner_printed) {
	  printf ("<!-- Translation from RTF performed by rtf2htm version %s\n",
		PROGRAM_VERSION);
	  printf (" !-- For information about this marvellous Open Source program\n");
	  printf (" !-- please go to http://recycle.50megs.com\n !-->\n");
	}
	banner_printed=TRUE;
}


/* Convert a two-char hexadecimal expression to an integer */
int
h2toi (char *s) {
	int tmp;
	int ch;
	tmp = tolower(*s++);
	if (tmp>'9') tmp-=('a'-10); 
	else tmp-='0';
	ch=16*tmp;
	tmp = tolower(*s++);
	if (tmp>'9') tmp-=('a'-10); 
	else tmp-='0';
	ch+=tmp;
	return ch;
}


char *word_string (Word *w) {
	char *str;
	CHECK_PARAM_NOT_NULL(w);
	if (w->hash_index) str = hash_get_string (w->hash_index);
	else str = NULL;
	return str;
}


Word * word_new (char *str) {
	Word * w;

	w = (Word *) malloc(sizeof(Word));
	if (!w)
		error_handler ("out of memory");
	bzero ((void*) w, sizeof(Word));
	if (!w) error_handler ("cannot allocate a Word");

	if (str)  w->hash_index = hash_get_index (str);
	else w->hash_index = 0;

	return w;
}



void word_free (Word *w) {
	Word *prev;
	Word *w2;

	return;

	CHECK_PARAM_NOT_NULL(w);

	while (w) {
		w2 = w->child;
		if (w2)
			word_free(w2);

		prev = w;
		w = w->next;
		free (prev);
	}
}


void
starting_body ()
{
	within_head=FALSE;

	if (!banner_printed)
		print_banner();

	if (!have_printed_body) {
		printf ("</head>\n<body>\n");
		within_head=FALSE;
		have_printed_body=TRUE;
	}
}


/*-------------------------------------------------------------------*/

typedef struct {
	int num;
	char *name;
} FontEntry;

#define MAX_FONTS (256)
static FontEntry font_table[MAX_FONTS];
static int total_fonts=0;

char* lookup_fontname (int num) {
	int i;
	if (total_fonts) 
	for(i=0;i<total_fonts;i++) {
		if (font_table[i].num==num) 
			return font_table[i].name;
	}
	return NULL;
}

void
process_font_table (Word *w)
{
	Word *w2;

	CHECK_PARAM_NOT_NULL(w);

	// Prevent text in Head from being printed.
	within_head=TRUE;

	while(w) {
		int num;
		char name[255];
		char *tmp;

		if ((w2=w->child)) {
			tmp = word_string (w2);
			if (!strncmp("\\f",tmp,2)) {
				num = atoi (&tmp[2]);
				name[0]=0;

				w2=w2->next;
				while(w2) {
					tmp = word_string (w2);
					if (tmp && tmp[0] != '\\')
						strcat(name,tmp);

					w2=w2->next;
				}

				if (debug_mode)
				  printf ("<!-- font table entry %d is %s -->\n", 
					num,name);

				// Chop the gall-derned semicolon.
				if ((tmp=strchr(name,';')))
					*tmp=0;

				font_table[total_fonts].num=num;
				font_table[total_fonts].name=strdup(name);
				total_fonts++;
			}
		}
		w=w->next;
	}

	printf ("<!-- font table contains %d fonts total-->\n",total_fonts);
	within_head=FALSE;
}


/*-------------------------------------------------------------------*/

/* Hardly an important group, since the data contained is often
 * inaccurate, but we may as well print it out.
 */

static char *month_strings[12]= {
#ifdef ENGLISH
  "January","February","March","April","May","June","July","August",
  "September","October","November","December"
#endif
#ifdef FRANCAIS
  "Janvier","Fevrier","Mars","Avril","Mai","Juin","Juillet","Aout","Septembre",
  "Octobre","Novembre","Decembre"
#endif
#ifdef ITALIANO
  "Ianuario","Febbraio","Marzo","Aprile","Maggio","Iuno",
  "Luglio","Agusto","Settembre","Ottobre","Novembre","Dicembre"
#endif
};

void word_dump_date(Word *w)
{
	int year=0, month=0, day=0, hour=0, minute=0;
	CHECK_PARAM_NOT_NULL(w);
	while (w) {
	 	char *s = word_string (w);
		if (*s == '\\') {
			++s;
			if (!strncmp (s, "yr", 2) && isdigit(s[2])) {
				year = atoi (&s[2]);
			}
			else if (!strncmp (s, "mo", 2) && isdigit(s[2])) {
				month= atoi (&s[2]);
			}
			else if (!strncmp (s, "dy", 2) && isdigit(s[2])) {
				day= atoi (&s[2]);
			}
			else if (!strncmp (s, "min", 3) && isdigit(s[3])) {
				minute= atoi (&s[3]);
			}
			else if (!strncmp (s, "hr", 2) && isdigit(s[2])) {
				hour= atoi (&s[2]);
			}
		}
		w=w->next;
	}
	if (year && month && day) {
		printf ("%d %s %d ", day, month_strings[month-1], year);
	}
	if (hour && minute) {
		printf ("%02d:%02d ", hour, minute);
	}
}



void
process_info_group (Word *w)
{
	Word *child;

	CHECK_PARAM_NOT_NULL(w);

	// Prevent text in Head from being printed.
	within_head=TRUE;

	while(w) {
		child = w->child;
		if (child) {
			Word *w2;
			char *s;

			s = word_string(child);

			if (!strcmp("\\title", s)) {
				printf ("<title>");
				w2=child->next;
				while (w2) {
					char *s2 = word_string(w2);
					if (s2[0] != '\\') printf ("%s ", s2);
					w2=w2->next;
				}
				printf ("</title>\n");
			} 
			else if (!strcmp("\\author", s)) {
				printf ("<!-- author: ");
				w2=child->next;
				while (w2) {
					char *s2 = word_string(w2);
					if (s2[0] != '\\') printf ("%s ", s2);
					w2=w2->next;
				}
				printf (" -->\n");
			} 
			else if (!strcmp("\\comment", s)) {
				printf ("<!-- comment: ");
				w2=child->next;
				while (w2) {
					char *s2 = word_string(w2);
					if (s2[0] != '\\') printf ("%s ", s2);
					w2=w2->next;
				}
				printf (" -->\n");
			}
			else if (!strncmp("\\nofpages", s, 9)) {
				printf ("<!-- total pages: %s -->\n",&s[9]);
			}
			else if (!strncmp("\\nofwords", s, 9)) {
				printf ("<!-- total words: %s -->\n",&s[9]);
			}
			else if (!strncmp("\\nofchars", s, 9) && isdigit(s[9])) {
				printf ("<!-- total chars: %s -->\n",&s[9]);
			}
			else if (!strncmp("\\edmins", s, 7)) {
				printf ("<!-- total editing minutes %s -->\n",
					&s[7]);
			}
			else if (!strcmp("\\creatim", s)) {
				printf ("<!-- creaton date: ");
				if (child->next) word_dump_date (child->next);
				printf (" -->\n");
			}
			else if (!strcmp("\\printim", s)) {
				printf ("<!-- last printed: ");
				if (child->next) word_dump_date (child->next);
				printf (" -->\n");
			}
			else if (!strcmp("\\buptim", s)) {
				printf ("<!-- last backup: ");
				if (child->next) word_dump_date (child->next);
				printf (" -->\n");
			}
			else if (!strcmp("\\revtim", s)) {
				printf ("<!-- revision date: ");
				if (child->next) word_dump_date (child->next);
				printf (" -->\n");
			}
		} 
		w = w->next;
	}
	within_head=FALSE;
}

/*-------------------------------------------------------------------*/

typedef struct {
	unsigned char r,g,b;
} Color;

#define MAX_COLORS (256)
static Color color_table[MAX_COLORS];
static int total_colors=0;

void
process_color_table (Word *w)
{
	int r,g,b;
	unsigned char flags=0;

	CHECK_PARAM_NOT_NULL(w);

	// Prevent text in Head from being printed.
	within_head=TRUE;

	while(w) {
		char *s = word_string (w);

#if 0
		printf ("found this color table word: %s\n", word_string(w));
#endif

		if(!strncmp("\\red",s,4)) {
			r = atoi(&s[4]);
			while(r>255) r>>=8;
			flags |= 1;
		}
		else if(!strncmp("\\green",s,6)) {
			g = atoi(&s[6]);
			while(g>255) g>>=8;
			flags |= 2;
		}
		else if(!strncmp("\\blue",s,5)) {
			b = atoi(&s[5]);
			while(b>255) b>>=8;
			flags |= 4;
		}

		/* Don't worry about the order */
		if (flags==7) {
			color_table[total_colors].r = r;
			color_table[total_colors].g = g;
			color_table[total_colors++].b = b;
			flags=0;
		}

		w=w->next;
	}

	if (debug_mode)
	  printf ("<!-- color table had %d entries -->\n", total_colors);

	within_head=FALSE;
}


/*-------------------------------------------------------------------*/

enum {
	ATTR_BOLD, ATTR_ITALIC, ATTR_UNDERLINE, ATTR_FONTSIZE, ATTR_FONTFACE,
	ATTR_FOREGROUND, ATTR_BACKGROUND, ATTR_MARKER, ATTR_CAPS,
	ATTR_DOT_UL, ATTR_DOUBLE_UL, ATTR_WORD_UL, ATTR_SHADOW,
	ATTR_SUPER, ATTR_SUB, ATTR_STRIKE, ATTR_OUTLINE, ATTR_EXPAND,
	ATTR_HIDDEN
};


/* XX convert to linked list in next version */
#define MAX_ATTRS (10000)
static unsigned char attr_stack [MAX_ATTRS];
static char *attr_stack_params [MAX_ATTRS];
static int tos=-1;

static void attr_push_core (int attr, char* param) {
	switch(attr) {
	case ATTR_BOLD: printf ("<b>"); break;
	case ATTR_ITALIC: printf ("<i>"); break;
	case ATTR_UNDERLINE: printf ("<u>"); break;
	case ATTR_FONTSIZE: 
		if (simple_mode) {
			int point_size=atoi (param);
			if (point_size<=8)
				printf ("<font size=1>");
			else if (point_size<=10)
				printf ("<font size=2>");
			else if (point_size<=12)
				printf ("<font size=3>");
			else if (point_size<=14)
				printf ("<font size=4>");
			else if (point_size<=18)
				printf ("<font size=5>");
			else if (point_size<=24)
				printf ("<font size=6>");
			else if (point_size<=36)
				printf ("<font size=7>");
			else 
				printf ("<font size=8>");
		}
		else
			printf ("<span style=\"font-size:%s\">",param); 
		break;
	case ATTR_FONTFACE: printf ("<font face=\"%s\">",param); break;
	case ATTR_FOREGROUND: printf ("<font color=%s>", param); break;
	case ATTR_BACKGROUND: printf ("<font bgcolor=%s>",param); break;
	case ATTR_DOT_UL: printf ("<span style=\"font-variant: \">"); break;
	case ATTR_DOUBLE_UL: printf ("<!-- begin double underline --><u>"); break;
	case ATTR_WORD_UL: printf ("<!-- begin word underline --><u>"); break;
	case ATTR_SUPER: printf ("<sup>"); break;
	case ATTR_SUB: printf ("<sub>"); break;
	case ATTR_STRIKE: printf ("<s>"); break;
	case ATTR_HIDDEN: printf ("<!-- "); break;
	case ATTR_OUTLINE: printf ("<!--begin outline-->"); break;
	case ATTR_SHADOW: printf ("<!--begin shadow-->"); break;
	case ATTR_EXPAND: printf ("<!--begin expand-->"); break;
	case ATTR_CAPS: printf ("<span style=\"font-variant: small-caps\">"); break;
	case ATTR_MARKER: break;
	}
}

static void attr_pop_core (int attr, char* param) {
	switch(attr) {
	case ATTR_BOLD: printf ("</b>"); break;
	case ATTR_ITALIC: printf ("</i>"); break;
	case ATTR_UNDERLINE: printf ("</u>"); break;
	case ATTR_FONTSIZE: 
		if (simple_mode)
			printf ("</font>"); 
		else
			printf ("</span>"); 
		break;
	case ATTR_FONTFACE: printf ("</font>"); break;
	case ATTR_FOREGROUND: printf ("</font>"); break;
	case ATTR_BACKGROUND: printf ("</font>"); break;
	case ATTR_DOT_UL: printf ("</u><!-- end dot underline -->"); break;
	case ATTR_DOUBLE_UL: printf ("</u><!-- end double underline -->"); break;
	case ATTR_WORD_UL: printf ("</u><!-- end word underline -->"); break;
	case ATTR_SUPER: printf ("</sup>"); break;
	case ATTR_SUB: printf ("</sub>"); break;
	case ATTR_STRIKE: printf ("</s>"); break;
	case ATTR_HIDDEN: printf (" -->"); break;
	case ATTR_OUTLINE: printf ("<!--end outline-->"); break;
	case ATTR_SHADOW: printf ("<!--end shadow-->"); break;
	case ATTR_EXPAND: printf ("<!--end expand-->"); break;
	case ATTR_CAPS: printf ("</span>"); break;
	case ATTR_MARKER: break;
	}
}

void attr_push(int attr, char* param) 
{
	if (tos>=MAX_ATTRS) { fprintf (stderr,"Too many attributes!\n"); return; }

	/* If the attribute refers to text, we're in the <body> section. */
	if (attr != ATTR_MARKER) {
		starting_body();
		starting_text();
	}

	++tos;
	attr_stack [tos]=attr;
	if (param) 
		attr_stack_params [tos]=strdup(param);
	else
		attr_stack_params [tos]=NULL;

	attr_push_core (attr, param);
}

void attr_push_mark() {
	attr_push(ATTR_MARKER,NULL);
}

int attr_pop(int attr) {
	if(tos>=0 && attr_stack[tos]==attr)
	{
		int attr = attr_stack [tos];
		char *param = attr_stack_params [tos];
		attr_pop_core (attr, param);
		if (param) free(param);
		tos--;
		return TRUE;
	}
	else
		return FALSE;
}

void attr_drop_all () {
	while (tos>=0 && attr_stack[tos]!=ATTR_MARKER) {
		char *param=attr_stack_params [tos];
		if (param) free(param);
		tos--;
	}
}

void attr_pop_mark() {
	if (!attr_pop(ATTR_MARKER))
		warning_handler ("cannot find ATTR_MARKER on attribute stack");
}

/* Routine to undo all attributes heretofore applied, reversing
 * the order in which they were applied.
 */
void attr_pop_all() {
	while (tos>=0 && attr_stack[tos]!=ATTR_MARKER) {
		int attr=attr_stack [tos];
		char *param=attr_stack_params [tos];
		if (param) free(param);
		attr_pop_core (attr, param);
		tos--;
	}
}

/* Routine to print out HTML to undo all attributes heretofore applied, 
 * but which actually leves it all on the stack. This is needed for \cell,
 * often occurs within a group, yet HTML uses <td></td> which in theory
 * will remove attributes at the end of that block.
 */
void attr_pop_dump() {
	int i=tos;
	while (i>=0 && attr_stack[i]!=ATTR_MARKER) {
		int attr=attr_stack [i];
		char *param=attr_stack_params [i];
		attr_pop_core (attr, param);
		i--;
	}
}

/* Routine to print out HTML to *redo* all attributes heretofore applied, 
 * but which actually leves it all on the stack. This is needed for \cell,
 * often occurs within a group, yet HTML uses <td></td> which in theory
 * will remove attributes at the end of that block.
 */
void attr_push_dump() {
	int i=tos;

	/* First find the last marker */
	while (i>=0 && attr_stack[i]!=ATTR_MARKER)
		i--;

	/* Now print everything above the marker */
	i++;
	while (i<=tos && attr_stack[i]!=ATTR_MARKER) {
		char *param = attr_stack_params [i];
		int attr = attr_stack[i++];
		attr_push_core (attr, param);
		i++;
	}
}

/*-------------------------------------------------------------------*/

void print_with_special_exprs (char *s) {
	int ch;
	CHECK_PARAM_NOT_NULL(s);
	while ((ch=*s++)) {
		if (ch=='<') printf ("&lt;");
		else if (ch=='>') printf ("&gt;");
		else if (ch=='&') printf ("&amp;");
		else printf ("%c",ch);
	}
}


static void begin_table()
{
	within_table=TRUE;
	have_printed_tr = FALSE;
	have_printed_td = FALSE;
	have_printed_end_tr = FALSE;
	have_printed_end_td = FALSE;
	attr_push_mark();
	starting_body();
	printf ("\n<table>\n");
}


void end_table ()
{
	if (within_table) {
		if (!have_printed_end_td) {
			attr_pop_dump();
			printf ("\n</td>\n");
		}
		if (!have_printed_end_tr) {
			printf ("\n</tr>\n");
		}
		printf ("\n</table>\n");
		within_table=FALSE;
		have_printed_tr = FALSE;
		have_printed_td = FALSE;
		have_printed_end_tr = FALSE;
		have_printed_end_td = FALSE;
	}
}


void starting_text() {
	if (within_table) {
		if (!have_printed_tr) {
			printf ("\n<tr>\n");
			have_printed_tr=TRUE;
			have_printed_end_tr=FALSE;
			have_printed_td=FALSE;
		}
		if (!have_printed_td) {
			printf ("\n<td>\n");
			attr_push_dump();
			have_printed_td=TRUE;
			have_printed_end_td=FALSE;
		}
	}
}

/* Recursive routine to print out HTML based on a tree of words.
 */
static void
word_print_html_core (Word *w, int paragraph_align)
{
	char *s;
	Word *prev_w=NULL;
	int within_picture=FALSE;
	char picture_path[255];
	int picture_width;
	int picture_height;
	int picture_bits_per_pixel=1;
#define PICT_UNKNOWN 0
#define PICT_WM 1
#define PICT_MAC 2
#define PICT_PM 3
#define PICT_DI 4
#define PICT_WB 5
	int picture_type=PICT_UNKNOWN;
 	int picture_wmetafile_type;
 	char *picture_wmetafile_type_str;
	int total_chars_this_line=0; /* for simulating \tab */
	FILE *f=NULL;
	int is_cell_group=FALSE;
	int new_paragraph_align=ALIGN_LEFT;

	CHECK_PARAM_NOT_NULL(w);

	if (!coming_pars_that_are_tabular && within_table) {
		end_table();
	}
	else if (coming_pars_that_are_tabular && !within_table) {
		begin_table();
	}

	switch (paragraph_align) {
	case ALIGN_CENTER:
		printf ("<center>"); break;
	case ALIGN_LEFT:
		break;
	case ALIGN_RIGHT:
		if (simple_mode)
			printf ("<div align=right>\n");
		else
			printf ("<div style=\"text-align:right\">\n"); 
		break;
	case ALIGN_JUSTIFY:
		if (simple_mode)
			printf ("<div align=justify>\n"); 
		else
			printf ("<div style=\"text-align:justify\">\n"); 
		break;
	}

	/* Mark our place in the stack */
	attr_push_mark();

	while (w) {
		s = word_string (w);

		if (s) {
			if (s[0] != '\\') {
				starting_body();

#if 0
				if (debug_mode)
				  printf ("<!-- have_printed_tr=%d  -->\n", have_printed_tr);
				if (debug_mode)
				  printf ("<!-- have_printed_td=%d  -->\n", have_printed_td);
#endif
				starting_text();

				if (within_picture) {
					starting_body();
					if (!f) {
						char *ext=NULL;
						switch (picture_type) {
						case PICT_WB: ext="bmp"; break;
						case PICT_WM: ext="wmf"; break;
						case PICT_MAC: ext="pict"; break;
						case PICT_DI: ext="dib"; break; /* Device independent bitmap=??? */
						case PICT_PM: ext="pmm"; break; /* OS/2 metafile=??? */
						}
						sprintf (picture_path, "pict%03d.%s", 
							picture_file_number++,ext);
						f=fopen(picture_path,"w");
					}
					if (s[0]!=' ') {
						char *s2;
						printf ("<!-- picture data found -->\n");
						if (picture_wmetafile_type_str) {
							printf ("<!-- Windows Metafile type is %s -->\n",
								picture_wmetafile_type_str);
						}
						printf ("<!-- picture dimensions: %d by %d, depth %d -->\n", 
							picture_width, picture_height, picture_bits_per_pixel);
						if (picture_width && picture_height && picture_bits_per_pixel) {
							s2=s;
							while (*s2) {
								unsigned int tmp,value;
								tmp=tolower(*s2++);
								if (tmp>'9') tmp-=('a'-10); 
								else tmp-='0';
								value=16*tmp;
								tmp=tolower(*s2++);
								if (tmp>'9') tmp-=('a'-10); 
								else tmp-='0';
								value+=tmp;
								fprintf (f,"%c", value);
							}
						}
					}
				}
				else {
					total_chars_this_line += strlen(s);
					print_with_special_exprs (s);
				}
/*----RTF keywords---------------------------------------------------------*/
			} else {
				s++;
				if (!strncmp (s, "rtf", 3)) {
					printf ("<html>\n<head>\n");
					print_banner();
				}
				else if (!strcmp (s, "ansi")) {
					charset_type = CHARSET_ANSI;
					printf ("<!-- document uses ANSI character set -->\n");
				}
				else if (!strcmp (s, "mac")) {
					charset_type = CHARSET_MAC;
					printf ("<!-- document uses Macintosh character set -->\n");
				}
				else if (!strcmp (s, "pc")) {
					charset_type = CHARSET_CP437;
					printf ("<!-- document uses PC codepage 437 character set -->\n");
				}
				else if (!strcmp (s, "pca")) {
					charset_type = CHARSET_CP850;
					printf ("<!-- document uses PC codepage 850 character set -->\n");
				}
				else if (!strcmp (s, "info")) {
					process_info_group (w->next);
					if (debug_mode)
					  printf("<!-- info group -->\n");
					break;
				}
				else if (!strcmp (s, "*")) {
					/* Various */
					if (debug_mode)
					  printf("<!-- destination skipped -->\n");
					break;
				}
/*----header/footer keywords------------------------------------------------*/
				else if (!strcmp (s, "header")) {
					break;
				}
				else if (!strcmp (s, "headerl")) {
					break;
				}
				else if (!strcmp (s, "headerr")) {
					break;
				}
				else if (!strcmp (s, "headerf")) {
					break;
				}
				else if (!strcmp (s, "footer")) {
					break;
				}
				else if (!strcmp (s, "footerl")) {
					break;
				}
				else if (!strcmp (s, "footerr")) {
					break;
				}
				else if (!strcmp (s, "footerf")) {
					break;
				}
/*----\pict keywords-------------------------------------------------------*/
				else if (!strcmp (s, "pict")) {
					printf ("<!-- picture begins -->\n");
					within_picture=TRUE;
					picture_width = picture_height = 0;
					picture_type = PICT_WB;
				}
				else if (!strncmp (s, "bin",3)) {
					if (debug_mode) printf ("<!-- picture is binary -->\n");
					warning_handler ("binary encoded picture found (hex is supported), skipped");
					break;
				}
				else if (!strcmp (s, "macpict")) {
					picture_type = PICT_MAC;
					if (debug_mode) printf ("<!-- Mac Quickdraw picture identifier -->\n");
				}
				else if (!strncmp (s, "pmmetafile",10)) {
					picture_type = PICT_PM;
					if (debug_mode) printf ("<!-- OS/2 metafile picture identifier -->\n");
				}
				else if (!strncmp (s, "wmetafile",9)) {
					picture_type = PICT_WM;
					if (debug_mode) printf ("<!-- Windows metafile picture identifier -->\n");
					if (within_picture) {
						char ch = s[9];
 						picture_wmetafile_type=ch;
						switch(ch) {
						case '1': picture_wmetafile_type_str="MM_TEXT"; break;
						case '2': picture_wmetafile_type_str="MM_LOMETRIC"; break;
						case '3': picture_wmetafile_type_str="MM_HIMETRIC"; break;
						case '4': picture_wmetafile_type_str="MM_LOENGLISH"; break;
						case '5': picture_wmetafile_type_str="MM_HIENGLISH"; break;
						case '6': picture_wmetafile_type_str="MM_TWIPS"; break;
						case '7': picture_wmetafile_type_str="MM_ISOTROPIC"; break;
						case '8': picture_wmetafile_type_str="MM_ANISOTROPIC"; break;
						default: picture_wmetafile_type_str="default:MM_TEXT"; break;
						}
					}
				}
				else if (!strncmp (s, "dibitmap",7)) {
					if (debug_mode) printf ("<!-- device independent picture identifier skipped -->\n");
				}
				else if (!strncmp (s, "wbitmap",7)) {
					if (debug_mode) printf ("<!-- bitmap picture identifier skipped -->\n");
				}
				else if (!strncmp (s, "wbmbitspixel",12)) {
					if (debug_mode) printf ("<!-- bitmap number bits per pixel -->\n");
					if (within_picture) {
						picture_bits_per_pixel = atoi(&s[12]);
					}
				}
				else if (!strncmp (s, "wbmplanes",9)) {
					if (debug_mode) printf ("<!-- bitmap number of planes skipped -->\n");
				}
				else if (!strncmp (s, "wbmwidthbytes",12)) {
					if (debug_mode) printf ("<!-- bitmap width in bytes skipped -->\n");
				}
				else if (!strncmp (s, "picwgoal",8)) {
					if (debug_mode) printf ("<!-- picture desired width %s skipped -->\n", &s[8]);
				}
				else if (!strncmp (s, "pichgoal",8)) {
					if (debug_mode) printf ("<!-- picture desired height %s skipped -->\n", &s[8]);
				}
				else if (!strncmp (s, "picw",4)) {
					if (debug_mode) printf ("<!-- bitmap/picture width %s (or xExt if metafile) -->\n", &s[4]);
					if (within_picture) {
						picture_width = atoi(&s[4]);
					}
				}
				else if (!strncmp (s, "pich",4)) {
					if (debug_mode) printf ("<!-- bitmap/picture height %s (or yExt if metafile) -->\n", &s[4]);
					if (within_picture) {
						picture_height = atoi(&s[4]);
					}
				}
				else if (!strncmp (s, "picscalex",9)) {
					if (debug_mode) printf ("<!-- picture horizontal scaling skipped -->\n");
				}
				else if (!strncmp (s, "pichgoal",8)) {
					if (debug_mode) printf ("<!-- picture vertical scaling skipped -->\n");
				}
				else if (!strcmp (s, "picscaled")) {
					if (debug_mode) printf ("<!-- picture (macpict) is scaled skipped -->\n");
				}
				else if (!strncmp (s, "piccropt", 8)) {
					if (debug_mode) printf ("<!-- picture top croping value skipped -->\n");
				}
				else if (!strncmp (s, "piccropb", 8)) {
					if (debug_mode) printf ("<!-- picture bottom croping value skipped -->\n");
				}
				else if (!strncmp (s, "piccropl", 8)) {
					if (debug_mode) printf ("<!-- picture left croping value skipped -->\n");
				}
				else if (!strncmp (s, "piccropr", 8)) {
					if (debug_mode) printf ("<!-- picture right croping value skipped -->\n");
				}
				else if (!strcmp (s, "brdrs")) {
					if (debug_mode) printf ("<!-- picture single border skipped -->\n");
				}
				else if (!strcmp (s, "brdrdb")) {
					if (debug_mode) printf ("<!-- picture double border skipped -->\n");
				}
				else if (!strcmp (s, "brdrth")) {
					if (debug_mode) printf ("<!-- picture thick border skipped -->\n");
				}
				else if (!strcmp (s, "brdrsh")) {
					if (debug_mode) printf ("<!-- picture shadow border skipped -->\n");
				}
				else if (!strcmp (s, "brdrdot")) {
					if (debug_mode) printf ("<!-- picture dotted border skipped -->\n");
				}
				else if (!strcmp (s, "brdrhair")) {
					if (debug_mode) printf ("<!-- picture hairline border skipped -->\n");
				}
				else if (!strncmp (s, "brdrw",5)) {
					if (debug_mode) printf ("<!-- picture's border's width skipped -->\n");
				}
				else if (!strncmp (s, "brdrcf",6)) {
					if (debug_mode) printf ("<!-- picture's border's color skipped -->\n");
				}
				else if (!strncmp (s, "shading",7)) {
					if (debug_mode) printf ("<!-- picture shading percentage skipped -->\n");
				}
				else if (!strcmp (s, "bghoriz")) {
					if (debug_mode) printf ("<!-- picture horizontal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgvert")) {
					if (debug_mode) printf ("<!-- picture vertical bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgfdiag")) {
					if (debug_mode) printf ("<!-- picture fwd diagonal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgbdiag")) {
					if (debug_mode) printf ("<!-- picture backward diagonal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgcross")) {
					if (debug_mode) printf ("<!-- picture crosshatched bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgdcross")) {
					if (debug_mode) printf ("<!-- picture diagonal crosshatched bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkhoriz")) {
					if (debug_mode) printf ("<!-- picture dark horizontal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkvert")) {
					if (debug_mode) printf ("<!-- picture dark vertical bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkfdiag")) {
					if (debug_mode) printf ("<!-- picture dark fwd diagonal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkbdiag")) {
					if (debug_mode) printf ("<!-- picture dark backward diagonal bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkcross")) {
					if (debug_mode) printf ("<!-- picture dark crosshatched bg pattern skipped -->\n");
				}
				else if (!strcmp (s, "bgkdcross")) {
					if (debug_mode) printf ("<!-- picture dark diagonal crosshatched bg pattern skipped -->\n");
				}
				else if (!strncmp (s, "cfpat",5)) {
					if (debug_mode) printf ("<!-- picture bg pattern's line color skipped -->\n");
				}
				else if (!strncmp (s, "cbpat",5)) {
					if (debug_mode) printf ("<!-- picture bg pattern's bg color skipped -->\n");
				}
/*----table of contents keywords-------------------------------------------*/
				else if (!strcmp (s, "tcf")) {
					if (debug_mode) printf ("<!-- TOC type skipped -->\n");
				}
				else if (!strcmp (s, "tcl")) {
					if (debug_mode) printf ("<!-- TOC level skipped -->\n");
				}
				else if (!strcmp (s, "tc")) {
					if (debug_mode) printf ("<!-- table of contents entry skipped -->\n");
				}
/*-- -table keywords-------------------------------------------------------*/
				else if (!strcmp (s, "cgrid")) {
					if (debug_mode) printf ("<!--  cgrid found  -->\n");
				}
				else if (!strcmp (s, "intbl")) {
					if (debug_mode) printf ("<!--  intbl found  -->\n");
					++coming_pars_that_are_tabular;
				}
				else if (!strncmp (s, "cellx", 5)) {
					if (debug_mode) printf ("<!-- cell position info  -->\n");
				}
				else if (!strcmp (s, "cell")) {
					is_cell_group=TRUE;
					if (!have_printed_td) {
						/* Need this with empty cells */
						printf ("\n<td>\n");
						attr_push_dump();
					}
					attr_pop_dump();
					printf ("\n</td>\n");
					have_printed_td = FALSE;
					have_printed_end_td=TRUE;
				}
				else if (!strcmp (s, "row")) {
					if (within_table) {
						printf ("\n</tr>\n");
						have_printed_tr = FALSE;
						have_printed_end_tr=TRUE;
					} else {
						if (debug_mode)
						  printf ("<!-- end of table row -->\n");
					}
				}
				else if (!strcmp (s, "trowd")) {
					if (debug_mode) printf ("<!-- table row defaults -->\n"); 
				}
				else if (!strncmp (s, "trgaph", 6)) {
					if (debug_mode) printf ("<!-- table inter-cell spacing  -->\n"); 
				}
				else if (!strcmp (s, "clmgf")) {
					if (debug_mode) printf ("<!-- table begin-cell-merge  -->\n"); 
				}
				else if (!strcmp (s, "clmrg")) {
					if (debug_mode) printf ("<!-- table merge-with-previous-cell  -->\n"); 
				}
				else if (!strncmp (s, "cellx", 5)) {
					if (debug_mode) printf ("<!-- table cell right-boundary change  -->\n"); 
				}
				else if (!strcmp (s, "clbrdrb")) {
					if (debug_mode) printf ("<!-- table cell bottom border  -->\n"); 
				}
				else if (!strcmp (s, "clbrdrt")) {
					if (debug_mode) printf ("<!-- table cell top border  -->\n"); 
				}
				else if (!strcmp (s, "clbrdrl")) {
					if (debug_mode) printf ("<!-- table cell left border  -->\n"); 
				}
				else if (!strcmp (s, "clbrdrr")) {
					if (debug_mode) printf ("<!-- table cell right border  -->\n"); 
				}
				else if (!strcmp (s, "trql")) {
					if (debug_mode) printf ("<!-- table row left justify  -->\n"); 
				}
				else if (!strcmp (s, "trqr")) {
					if (debug_mode) printf ("<!-- table row right justify  -->\n"); 
				}
				else if (!strcmp (s, "trqc")) {
					if (debug_mode) printf ("<!-- table row center justify  -->\n"); 
				}
				else if (!strncmp (s, "trleft", 6)) {
					if (debug_mode) printf ("<!-- table row left position  -->\n"); 
				}
				else if (!strncmp (s, "trrh", 4)) {
					if (debug_mode) printf ("<!-- table row head size  -->\n"); 
				}
				else if (!strncmp (s, "clshdng", 7)) {
					if (debug_mode) printf ("<!-- table cell shading  -->\n"); 
				}
				else if (!strcmp (s, "clbghoriz")) {
					if (debug_mode) printf ("<!-- table cell horizontal bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgvert")) {
					if (debug_mode) printf ("<!-- table cell vertical bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgfdiag")) {
					if (debug_mode) printf ("<!-- table cell fwd diagonal bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbdiag")) {
					if (debug_mode) printf ("<!-- table cell backwd diagonal bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgdcross")) {
					if (debug_mode) printf ("<!-- table cell diagonal crosshatch bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbcross")) {
					if (debug_mode) printf ("<!-- table cell crosshatch bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkhor")) {
					if (debug_mode) printf ("<!-- table cell dark horizontal bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkvert")) {
					if (debug_mode) printf ("<!-- table cell dark vertical bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkfdiag")) {
					if (debug_mode) printf ("<!-- table cell dark fwd diag bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkbdiag")) {
					if (debug_mode) printf ("<!-- table cell dark backwd diag bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkcross")) {
					if (debug_mode) printf ("<!-- table cell dark crosshatch bg pattern  -->\n"); 
				}
				else if (!strcmp (s, "clbgbkdcross")) {
					if (debug_mode) printf ("<!-- table cell dark diag crosshatch bg pattern  -->\n"); 
				}
				else if (!strncmp (s, "clcfpat", 7)) {
					if (debug_mode) printf ("<!-- table cell background line-pattern line color  -->\n"); 
				}
				else if (!strncmp (s, "clcbpat", 7)) {
					if (debug_mode) printf ("<!-- table cell background line-pattern bg color  -->\n"); 
				}
/*----styles----------------------------------------------------------------*/
				else if (!strcmp (s, "stylesheet")) {
					/* Definition of Styles */
					if (debug_mode)
					  printf("<!-- style sheet -->\n");
					break;
				}
				else if (s[0]=='s' && isdigit(s[1])) {
					int stylenum = atoi(&s[1]);
					if (debug_mode)
					  printf ("<!-- style %d change ignored -->\n", stylenum);
				}
/*----section formatting----------------------------------------------------*/
				else if (!strcmp (s, "sect")) {
					/* section and paragraph break */
					printf ("<p>\n");
				}
				else if (!strcmp (s, "column")) {
					/* column break */

					/* Impossible to implement since
					 * there is no concept of "page length"
					 * in HTML to allow "snaking" of text
					 * from column to column.
					 */
					if (debug_mode) 
					  printf("<!-- column break ignored-->\n");
				}
				else if (!strncmp (s, "cols", 4) && 
				  isdigit (s[4])) {
					int num=atoi (&s[4]);
					if (debug_mode) 
					  printf("<!-- multiple columns (%d) request ignored-->\n",num);
				}
/*----paragraph formatting--------------------------------------------------*/
				else if (!strcmp (s, "par")) {
					printf ("<br>\n");
				}
				else if (!strcmp (s, "line")) {
					printf ("<br>\n");
				}
				else if (!strcmp (s, "page")) {
					printf ("<hr><hr>\n");
				}
				else if (!strcmp (s, "pard")) {
					attr_pop_all();
					if (debug_mode)
					  printf ("<!-- pard -->\n");
					if(coming_pars_that_are_tabular) {
						--coming_pars_that_are_tabular;
					}
				}
				else if (!strcmp (s, "ql")) {
					new_paragraph_align = ALIGN_LEFT;
					if (debug_mode)
					  printf ("<!-- left align ignored -->\n");
				}
				else if (!strcmp (s, "qr")) {
					new_paragraph_align = ALIGN_RIGHT;
					if (debug_mode)
					  printf ("<!-- right align ignored -->\n");
				}
				else if (!strcmp (s, "qj")) {
					new_paragraph_align = ALIGN_JUSTIFY;
					if (debug_mode)
					  printf ("<!-- justify ignored -->\n");
				}
				else if (!strcmp (s, "qc")) {
					new_paragraph_align = ALIGN_CENTER;
					if (debug_mode)
					  printf ("<!-- center ignored -->\n");
				}
				else if (!strncmp (s, "fi", 2) && isdigit(s[2])) {
					if (debug_mode)
					  printf ("<!-- first-line indent ignored -->\n");
				}
				else if (!strncmp (s, "li", 2) && isdigit(s[2])) {
					if (debug_mode)
					  printf ("<!-- left indent ignored -->\n");
				}
				else if (!strncmp (s, "ri", 2) && isdigit(s[2])) {
					if (debug_mode)
					  printf ("<!-- right indent ignored -->\n");
				}
/*----special chars---------------------------------------------------------*/
				else if (!strcmp (s, "~")) {
					printf ("&nbsp;");
				}
				else if (!strcmp (s, "emdash")) {
					printf ("&mdash;");
				}
				else if (!strcmp (s, "endash")) {
					printf ("&ndash;");
				}
				else if (!strcmp (s, "-")) {
					printf ("-"); /* optional hyphen */
				}
				else if (!strcmp (s, "_")) {
					printf ("-"); /* nonbreaking hyphen */
				}
				else if (!strcmp (s, "bullet")) {
					printf ("&bull;");
					/* It's arguable that this should be converted into
				 	 * an <li> but for now this will do.
					 */
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "ldblquote")) {
					printf ("&ldquo;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "rdblquote")) {
					printf ("&rdquo;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "lquote")) {
					printf ("&lsquo;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "rquote")) {
					printf ("&rsquo;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strncmp (s, "'", 1)) {
					/* \'XX is a hex char code expression */
					int ch = h2toi (&s[1]);
					if (ch >= 32 && ch < 127) {
						putchar (ch);
					} if (ch >= 160) {
						char *s2;
						switch (charset_type) {
						case CHARSET_ANSI:
							s2=special_exprs_ansi [ch-128];
							break;
						case CHARSET_MAC:
							s2=special_exprs_mac [ch-128];
							break;
						case CHARSET_CP437:
							s2=special_exprs_cp437 [ch-128];
							break;
						case CHARSET_CP850:
							s2=special_exprs_cp850 [ch-128];
							break;
						}
						if (!s2 || !*s2) 
							printf("<!--0x%02x-->",ch);
						else
							printf ("&%s;", s2);
					}
				}
/* 0xa9=ANSI, ?=codepage437 */	else if (!strcmp (s, "\\'a9")) {
					printf ("&copy;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "\\'ae")) {
					printf ("&reg;");
					++total_chars_this_line; /* \tab */
				}
				else if (!strcmp (s, "tab")) {
					/* Tab presents a genuine problem
					 * since HTML doesn't have an equivalent.
					 * As a kludge fix, I shall assume
					 * the font is fixed width and that
					 * the tabstops are 8 characters apart.
					 */
					int need= 8-(total_chars_this_line%8);
					total_chars_this_line += need;
					while(need>0) {
						printf ("&nbsp;");
						need--;
					}
					printf ("\n");
				}
/*----font attribute keywords-----------------------------------------------*/
				else if (!strcmp (s, "plain")) {
					attr_pop_all();
				}
				else if (!strcmp (s, "fnil")) {
					attr_push(ATTR_FONTFACE,FONTNIL_STR);
				}
				else if (!strcmp (s, "froman")) {
					attr_push(ATTR_FONTFACE,FONTROMAN_STR);
				}
				else if (!strcmp (s, "fswiss")) {
					attr_push(ATTR_FONTFACE,FONTSWISS_STR);
				}
				else if (!strcmp (s, "fmodern")) {
					attr_push(ATTR_FONTFACE,FONTMODERN_STR);
				}
				else if (!strcmp (s, "fscript")) {
					attr_push(ATTR_FONTFACE,FONTSCRIPT_STR);
				}
				else if (!strcmp (s, "fdecor")) {
					attr_push(ATTR_FONTFACE,FONTDECOR_STR);
				}
				else if (!strcmp (s, "fdecor")) {
					attr_push(ATTR_FONTFACE,FONTTECH_STR);
				}
				else if (!strcmp (s, "b")) {
					attr_push(ATTR_BOLD,NULL);
				}
				else if (!strcmp (s, "b0")) {
					attr_pop(ATTR_BOLD);
				}
				else if (!strcmp (s, "i")) {
					attr_push(ATTR_ITALIC,NULL);
				}
				else if (!strcmp (s, "i0")) {
					attr_pop(ATTR_ITALIC);
				}
				else if (!strcmp (s, "v")) {
				        attr_push(ATTR_HIDDEN,NULL);
				}
				else if (!strcmp (s, "super")) {
				        attr_push(ATTR_SUPER,NULL);
				}
				else if (!strcmp (s, "sub")) {
				        attr_push(ATTR_SUB,NULL);
				}
				else if (!strncmp (s, "expand", 6)) {
					int amount = atoi (&s[6]);
					if (!amount) attr_pop(ATTR_EXPAND);
					else attr_push(ATTR_EXPAND, &s[6]);
				}
				else if (!strcmp (s, "outl")) {
					attr_push(ATTR_OUTLINE,NULL);
				}
				else if (!strcmp (s, "outl0")) {
					attr_pop(ATTR_OUTLINE);
				}
				else if (!strcmp (s, "caps")) {
					attr_push(ATTR_CAPS,NULL);
				}
				else if (!strcmp (s, "caps0")) {
					attr_pop(ATTR_CAPS);
				}
				else if (!strcmp (s, "shad")) {
					attr_push(ATTR_SHADOW,NULL);
				}
				else if (!strcmp (s, "shad0")) {
					attr_pop(ATTR_SHADOW);
				}
				else if (!strncmp (s, "up",2) && isdigit(s[2])) {
					if (s[2]=='0')
						attr_pop(ATTR_SUPER);
					else
						attr_push(ATTR_SUPER,NULL);
				}
				else if (!strncmp (s, "dn",2) && isdigit(s[2])) {
					if (s[2]=='0')
						attr_pop(ATTR_SUB);
					else
						attr_push(ATTR_SUB,NULL);
				}
				else if (!strcmp (s, "strike")) {
					attr_push(ATTR_STRIKE,NULL);
				}
				else if (!strcmp (s, "strike0")) {
					attr_pop(ATTR_STRIKE);
				}
				else if (!strcmp (s, "ul")) {
					attr_push(ATTR_UNDERLINE,NULL);
				}
				else if (!strcmp (s, "uld")) {
				        attr_push(ATTR_DOT_UL,NULL);
				}
				else if (!strcmp (s, "uldb")) {
					attr_push(ATTR_DOUBLE_UL,NULL);
				}
				else if (!strcmp (s, "ulw")) {
					attr_push(ATTR_WORD_UL,NULL);
				}
				else if (!strcmp (s, "ulnone")) {
					if (!attr_pop(ATTR_UNDERLINE))
						;
				}
				else if (s[0]=='f' && s[1]=='s' && !s[2]) {
					/* \fs ## is erroneous AFAIK */
				}
				else if (s[0]=='f' && s[1]=='s' && isdigit(s[2])) {
					char str[20];
					int points = atoi(&s[2]);
					/* Note, fs20 means 10pt */
					points /= 2;
					sprintf (str,"%d",points);
					strcat (str,"pt");
					attr_push(ATTR_FONTSIZE,str);
				}
				else if (s[0]=='f' && isdigit(s[1])) {
					int num; 
					char *name;
					num = atoi (&s[1]);;
					name = lookup_fontname(num);
					if(!name) {
						printf ("<!-- invalid font number %d -->\n",num);
					} else {
						attr_push(ATTR_FONTFACE,name);
					}
				}
				else if (!strncmp (s, "colortbl", 8)) {
					/* Fonts used in document */
					if (debug_mode)
					  printf("<!-- color table -->\n");
					if (w->next) {
						process_color_table(w->next);
					}
					break;
				}
				else if (s[0]=='c' && s[1]=='f' && isdigit(s[2])) {
					int num = atoi (&s[2]);
					char str[40];
					if (num>=total_colors) {
						printf ("<! -- font color change attempted which is invalid %d  -->\n", num);
					}
					else
					{
						sprintf (str,"#%02x%02x%02x",
							color_table[num].r,
							color_table[num].g,
							color_table[num].b);
						attr_push(ATTR_FOREGROUND,str);
					}
				}
				else if (!strncmp (s, "fonttbl", 7)) {
					/* Fonts used in document */
					if (debug_mode)
					  printf("<!-- font table -->\n");
					if (w->next) {
						process_font_table(w->next);
					}
					break;
				}
/*-------------------------------------------------------------------------*/
/*-------------------------------------------------------------------------*/
			}
		} else {
			Word *child = w->child;

			if (child)
			  word_print_html_core (child, new_paragraph_align);
		}

		prev_w = w;
		w = w->next;
	}

	switch (paragraph_align) {
	case ALIGN_CENTER:
		printf ("</center>"); break;
	case ALIGN_LEFT:
		break;
	case ALIGN_RIGHT:
	case ALIGN_JUSTIFY:
		printf ("</div>\n"); 
		break;
	}

	if (within_picture) {
		if(f) { 
			fclose(f);
			printf ("<img src=%s>\n", picture_path);
			printf ("<!---end of picture -->\n");
		}
	}

	/* Undo font attributes UNLESS we're doing table cells
	 * since they would appear between </td> and </tr>.
	 */
	if (!is_cell_group)
		attr_pop_all();
	else
		attr_drop_all();

	attr_pop_mark();
}



void word_print_html(Word *w) 
{
	CHECK_PARAM_NOT_NULL (w);

	within_head=TRUE;
	have_printed_body=FALSE;
	within_table=FALSE;
	word_print_html_core (w, ALIGN_LEFT);
	end_table();
	printf ("\n</body></html>\n");
}
